// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef R_MATH_H_
#define R_MATH_H_

/*###########################################################################*/
#include <algorithm>
#include <cmath>
#include "RVector.h"
#include "RStd.h"
#include "RGen.h"
template <typename T> class RVector;
class ostream;

/*###########################################################################*/
unsigned gcd(const RVector<unsigned> &num);
double gcd(const RVector<double> &num, double slack = 5.0);

unsigned lcm(const RVector<unsigned> &num);
double lcm(const RVector<double> &num, double slack = 5.0);

// Ring modulus.
template <typename T, typename P> T ring_mod(T a, P period);
template <typename P> unsigned ring_mod(unsigned a, P period);
template <typename P> unsigned long ring_mod(unsigned long a, P period);
template <typename P> double ring_mod(double a, P period);
template <typename P> float ring_mod(float a, P period);

template <typename T, typename P> T interval_round(T a, P period);
template <typename P> double interval_round(double a, P period);
template <typename P> float interval_round(float a, P period);

template<typename D> D clip(const D &x, D low, D high)
  {return min(max(x, low), high);}

/*===========================================================================*/
bool boltzmann_trial(double old_cost, double new_cost, double temperature,
  RGen &rgen);

/*===========================================================================*/
class MathVec : public RVector<double> {
	typedef RVector<double> super;

public:
	MathVec() {}

	MathVec(size_type n, const double &value) : super(n, value) {}
	MathVec(int n, const double &value) : super(n, value) {}
	MathVec(long n, const double &value) : super(n, value) {}
	explicit MathVec(size_type n) : super(n) {}
	MathVec(const vector<double> &x) : super(x) {}

	template <typename InputIterator>
	  MathVec(InputIterator first, InputIterator last) :
	  super(first, last) {}

	MathVec &operator+=(const MathVec &a);
	MathVec &operator+=(const double &a);
	MathVec &operator-=(const MathVec &a);
	MathVec &operator-=(const double &a);
	MathVec &operator/=(const MathVec &a);
	MathVec &operator/=(const double &a);
	MathVec &flat_mul(const MathVec &a);
	MathVec &flat_mul(const double &a);
	MathVec &random_dir(double magnitude = 1.0, RGen &rg = RGen::gen);
	MathVec &normalize(double magnitude = 1.0);
	MathVec &clip(double low, double high);
	double mag() const;
	void print_to(ostream &os) const;
};

inline MathVec operator+(const MathVec &a, const MathVec &b)
  {return MathVec(a) += b;}
inline MathVec operator+(const MathVec &a, const double &b)
  {return MathVec(a) += b;}
inline MathVec operator-(const MathVec &a, const MathVec &b)
  {return MathVec(a) -= b;}
inline MathVec operator-(const MathVec &a, const double &b)
  {return MathVec(a) -= b;}
inline MathVec operator/(const MathVec &a, const MathVec &b)
  {return MathVec(a) /= b;}
inline MathVec operator/(const MathVec &a, const double &b)
  {return MathVec(a) /= b;}
inline MathVec flat_mul(const MathVec &a, const MathVec &b)
  {return MathVec(a).flat_mul(b);}
inline MathVec flat_mul(const double &a, const double &b)
  {return MathVec(a).flat_mul(b);}
double operator*(const MathVec &a, const MathVec &b);
RVector<double> operator*(const RVector<MathVec> &a, const MathVec &b);
inline ostream &operator<<(ostream &os, const MathVec &a)
  {a.print_to(os); return os;}
MathVec max_dimen(RVector<MathVec> &vec);
MathVec min_dimen(RVector<MathVec> &vec);
RVector<MathVec> &normalize(RVector<MathVec> &vec);

/*===========================================================================*/
void RMath_test();

/*###########################################################################*/
#include "RMath.cct"
#endif
