// Copyright 2000 by Robert Dick.
// All rights reserved.

#include <cctype>
#include <algorithm>
#include <functional>
#include <strstream>
#include <cstring>
#include "RStd.h"
#include "RString.h"
#include "RAlgo.h"

/*###########################################################################*/
RString::RString(bool a) {
	const char *buf = a ? "true" : "false";
	insert(end(), buf, buf + strlen(buf) + 1);
}

/*===========================================================================*/
RString::RString(char a) :
	data_(1, a)
{
	data_.push_back('\0');
}

/*===========================================================================*/
template <typename T>
void RString::init_from(T a) {
	ostrstream oss;
	oss << a << ends;
	data_.insert(data_.end(), oss.str(), oss.str() + oss.pcount());
}

/*===========================================================================*/
RString::RString(int a) {init_from(a);}
RString::RString(unsigned a) {init_from(a);}
RString::RString(long a) {init_from(a);}
RString::RString(unsigned long a) {init_from(a);}
RString::RString(double a) {init_from(a);}

/*===========================================================================*/
RString &RString::operator+=(const RString &a) {
	insert(end(), a.begin(), a.end());
	return *this;
}

/*===========================================================================*/
bool RString::append_line(istream &is, char terminator) {
	strstreambuf buf;
	is.get(buf, terminator);
	is.get();

	if (! is)
		return false;

	buf.sputc(0);
	*this += RString(buf.str());

 	return true;
}

/*===========================================================================*/
RVector<RString> RString::
tokenize(const char *divider) const {
	RVector<RString> vec;
	RString empty;
	pair<RString, const_iterator> result(empty, begin());

	while (result.second != end()) {
		result = token(result.second, divider);
		if (result.first.size())
			vec.push_back(result.first);
	}

	return vec;
}

/*===========================================================================*/
pair<RString, RString::iterator> RString::
token(const_iterator first, const char *divider) {
	pair<RString, RString::const_iterator> result =
	  const_cast<const RString *>(this)->token(first, divider);

	return
	  make_pair(result.first, const_cast<RString::iterator>(result.second));
}

/*===========================================================================*/
pair<RString, RString::const_iterator> RString::
token(const_iterator first, const char *divider) const {
// Walk past divider characters.
	const char *divider_end = divider + strlen(divider);

	while (first != end() &&
		find(divider, divider_end, *first) != divider_end)
	{
		first++;
	}

	if (first == end())
		return make_pair(RString(), end());

	const_iterator tail = find_first_of(first, end(), divider,
	  divider_end);

	return make_pair(RString(first, tail), tail);
}

/*===========================================================================*/
pair<RString, RString> RString::cut_tail(const char *divider = "/") const {
// Find the first match.
	RVector<char>::const_reverse_iterator ri =
	  find_first_of(rbegin(), rend(), divider, divider + strlen(divider));

	pair<RString, RString> ret_val;

	if (ri == rend()) {
// No dividers.
		ret_val.second = *this;
	} else {
		ret_val.first = RString(begin(), ri.base());
		ret_val.second = RString(ri.base(), end());
	}

	return ret_val;
}

/*===========================================================================*/
struct char_insen_less : public binary_function<char, char, bool> {
	bool operator()(char x, char y) {return tolower(x) < tolower(y);}
};

/*===========================================================================*/
struct char_insen_equal_to : public binary_function<char, char, bool> {
	bool operator()(char x, char y) {return tolower(x) == tolower(y);}
};

/*===========================================================================*/
bool RString::insen_less::operator()(const RString &a,
const RString &b) const {
	return lexicographical_compare(a.begin(), a.end(), b.begin(), b.end(),
	  char_insen_less());
}

/*===========================================================================*/
bool RString::insen_equal_to::
operator()(const RString &a, const RString &b) const {
	return a.size() == b.size() && equal(a.begin(), a.end(), b.begin(),
	  char_insen_equal_to());
}

/*===========================================================================*/
void RString::push_back(const_reference a) {
	data_.back() = a;
	data_.push_back(0);
}

/*===========================================================================*/
void RString::test() {
	RString str_a = "Fred";
	RString str_b = str_a;

	Rassert(str_a == str_b);
	Rassert(str_b == "Fred");

	str_b[0] = 'n';
	Rassert(str_b == "nred");

	str_a = "Elmo";
	Rassert(str_a == "Elmo");

	int x;
	x = RString("5").convert();
	Rassert(x == 5);

	long y;
	y = RString("-24").convert();
	Rassert(y == -24);

	RVector<RString> vec =
	  RString("I was walking down the street.").tokenize();

	Rassert(vec[0] == "I");
	Rassert(vec[1] == "was");
	Rassert(vec[2] == "walking");
	Rassert(vec[3] == "down");
	Rassert(vec[4] == "the");
	Rassert(vec[5] == "street.");
	Rassert(vec[4] == "the");

	RString full_path = "/usr/local/bin/fred";
	pair<RString, RString> cut_path = full_path.cut_tail();
	Rassert(cut_path.first == RString("/usr/local/bin/"));
	Rassert(cut_path.second == RString("fred"));

//	RString voltage = "12";
//	double volt = voltage.convert();
//	cout << volt << "\n";
}

/*===========================================================================*/
RString::Converter::operator bool() const {
	insen_equal_to i_eq;
	return (i_eq(ref_, "no") ||
	  i_eq(ref_, "0") ||
	  i_eq(ref_, "false") ||
	  i_eq(ref_, "f") ||
	  i_eq(ref_, "n")) ?
	  false : true;
}

/*===========================================================================*/
template <typename T> T RString::Converter::
convert_to() const {
	double a;
	strstream ss;
	ss << ref_;
	ss >> a;
	Rassert(! ss.fail());
	return T(a);
}

RString::Converter::operator int() const {return convert_to<int>();}
RString::Converter::operator unsigned() const {return convert_to<unsigned>();}
RString::Converter::operator long() const {return convert_to<long>();}
RString::Converter::operator unsigned long() const
  {return convert_to<unsigned long>();}
RString::Converter::operator float() const {return convert_to<float>();}
RString::Converter::operator double() const {return convert_to<double>();}
