// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

#include "TTdag.h"

#include <iomanip.h>
#include <algorithm>
#include <cmath>

#include "RMath.h"
#include "Epsilon.h"
#include "RVector.h"
#include "RStd.h"
#include "RGen.h"
#include "ArgPack.h"

/*###########################################################################*/
TTnode::TTnode(void) :
	type(-1),
	name(-1),
	deadline(0.0),
	hard(true),
	attrib_(ArgPack::task_attrib_av_.size(), HUGE_VAL)
{}

/*===========================================================================*/
TTnode::TTnode(int i) :
	type(-1),
	name(i),
	deadline(0.0),
	hard(true)
{
	MAP(x, ArgPack::task_attrib_av_.size()) {
		double val = ArgPack::task_attrib_av_[x] +
		  RGen::gen.flat_range_d(-ArgPack::task_attrib_mul_[x],
		  ArgPack::task_attrib_mul_[x]);

		if (eps_not_equal_to<double>()(ArgPack::task_attrib_round_[x], 0.0)) {
			val = interval_round(val, ArgPack::task_attrib_round_[x]);
		}

		attrib_.push_back(val);
	}
}

/*===========================================================================*/
TTnode::TTnode(int i, int t, unsigned ri, unsigned ru) :
	type(t),
	ri_(ri),
	ru_(ru),
	name(i),
	deadline(0.0)
{
	MAP(x, ArgPack::task_attrib_av_.size()) {
		double val = ArgPack::task_attrib_av_[x] +
		  RGen::gen.flat_range_d(-ArgPack::task_attrib_mul_[x],
		  ArgPack::task_attrib_mul_[x]);

		if (eps_not_equal_to<double>()(ArgPack::task_attrib_round_[x], 0.0)) {
			val = interval_round(val, ArgPack::task_attrib_round_[x]);
		}

		attrib_.push_back(val);
	}
}

/*###########################################################################*/
TTdag::TTdag() :
	low_bound_(-1),
	in_deg_(-1),
	out_deg_(-1),
	period_(-1.0)
{}

/*===========================================================================*/
void TTdag::init(int number, int in_deg, int out_deg, double period) {
	in_deg_ = in_deg;
	out_deg_ = out_deg;
	number_ = number;
	period_ = period;
}
/*===========================================================================*/
void TTdag::init(int low_bound, int in_deg, int out_deg, int number,
int first_task) {
	low_bound_ = low_bound;
	in_deg_ = in_deg;
	out_deg_ = out_deg;
	number_ = number;
	first_task_ = first_task;

	while (NumNodes() < low_bound_)
		augment();

	compute_dims();

	deadlines();
}

/*===========================================================================*/
void TTdag::augment() {
	if (! NumNodes()) {		// first augment, add single node
		TTnode *nd = new TTnode(0);
		AddNode(nd);
		return;
	}

// SELECT TEMPLATE (either IN or OUT now)

	if (RGen::gen.flip()) {
// Pick FAN-OUT
		SubListman	nodes_with_most_fanout;
		int			max_fanout_found = 0;

		DAG_FOREACH_NODE(TTnode, n, this)
			int node_fanout_avail = out_deg_ - NumOutArcs(n);
			if( node_fanout_avail <= 0 ) continue;
			if( node_fanout_avail > max_fanout_found ) {
				max_fanout_found = node_fanout_avail;
				nodes_with_most_fanout.Empty();
				nodes_with_most_fanout.InsertAfter(0,n);
			}
			else if( node_fanout_avail == max_fanout_found ) {
				nodes_with_most_fanout.InsertAfter(0,n);
			}
		DAG_ENDEACH_NODE

		int nnodes = nodes_with_most_fanout.NumOnList();
		if( nnodes > 0 ) {
			// at least one has fanout room, and the list is those with
			// the most room, randomly select one
			int m = RGen::gen.flat_range_l(0, nnodes);

			// pick node
			TTnode	*ap = (TTnode *)nodes_with_most_fanout[m]->data;
			int node_fanout_avail = out_deg_ - NumOutArcs(ap);

			// pick actual fanout value, including 0 as a bound
			int n = RGen::gen.flat_range_l(0, node_fanout_avail + 1);

			// add in the new nodes and arcs
			for (int x=0; x < n; x++) {
				TTnode	*node = new TTnode(NumNodes());
				AddNode(node);

				TTarc	*arc = new TTarc(NumArcs());
				AddArc(arc, ap, node);
			}
		}
	}
	else {
// Pick FAN-IN
		// find out how many existing nodes are not over their
		// OUTDEG limit...
		int	nf = 0;
		DAG_FOREACH_NODE(TTnode, n, this)
			if (NumOutArcs(n) < out_deg_)
				nf++;
		DAG_ENDEACH_NODE

// at least one
		RASSERT(nf);

		// determine # in deg
		int o = min(NumNodes(), min(in_deg_, nf));
		int n = RGen::gen.flat_range_l(1, o + 1);

		// create new node
		TTnode	*node = new TTnode(NumNodes());
		AddNode(node);

		// find places to attach, ensure no dupl arcs
		TTnodeSubList *ttl = new TTnodeSubList;
		DAG_FOREACH_NODE(TTnode, x, this)
			// use "h" field as temporary
			x->h = (x == node ? 1 : 0);
		DAG_ENDEACH_NODE

		RVector<int> index(NumNodes());
		MAP(x, index.size())
			index[x] = x;
		random_shuffle(index.begin(), index.end(), RGen::gen);

		unsigned long i = 0;
		int c = 0;
		while (c < o) {
			// randomly pick node, make sure not used before
			RASSERT(i < index.size());
			long elt_num = index[i++];

			TTnode *trynode = TTdag::NodeElement(elt_num);

			if (trynode->h || NumOutArcs(trynode) >= out_deg_)
				continue;

			// there is some danger here in that if the random #
			// generator fails to eventually pick all values that
			// we could loop forever (good thing we use a good rand).

			TTarc *arc = new TTarc(NumArcs());
			AddArc(arc,trynode,node);
			c++;
		}
		delete ttl;
	}
}

/*===========================================================================*/
void TTdag::print_to(ostream &os) const {
	if (ArgPack::period_mul_.size())
		os << "\tPERIOD " << period_ << "\n\n";
	else
		os << "\tAPERIODIC\n\n";

	int index = first_task_;
	DAG_FOREACH_NODE(TTnode, x, this)
		os << "\tTASK " << gen_name("t", x->name) << "\tTYPE ";

		if (x->type != -1)
			os << x->type << "\n";
		else if (ArgPack::task_unique_)
			os << index++ << "\n";
		else
			os << RGen::gen.flat_range_l(0, ArgPack::task_type_cnt_) << " ";

		MAP(y, ArgPack::task_attrib_name_.size()) {
			os << ArgPack::task_attrib_name_[y] << " " << x->attrib_[y] <<
			  " ";
		}
		os << "\n";
	DAG_ENDEACH_NODE

	os << "\n";

	DAG_FOREACH_ARC(TTarc, a, this)
		os << "\tARC " << setw(6) << gen_name("a", a->name) << " FROM " <<
		  setw(6) << gen_name("t", FromNode(a)->name) << " TO " <<
		  setw(6) << gen_name("t", ToNode(a)->name);

		int task_type;
		if( a->type < 0 )
			task_type = RGen::gen.flat_range_l(0, ArgPack::trans_type_cnt_);
		else
			task_type = a->type;

		os << " TYPE " << task_type << endl;
	DAG_ENDEACH_ARC

	os << "\n";

// Deadlines
// SIMPLE FOR NOW: PUT DEADLINE ON ALL "TERMINAL PROCESSES"
	int dcount = 0;
	DAG_FOREACH_NODE(TTnode, x, this)
		if(x->deadline > 0) {
			if (x->hard) {
				os	<< "\tHARD_DEADLINE " <<
				  setw(6) << gen_name("d", dcount++)  <<
				  " ON " << setw(6) << gen_name("t", x->name) << " AT " <<
				  x->deadline << "\n";
			} else {
				os	<< "\tSOFT_DEADLINE " <<
				  setw(6) << gen_name("d", dcount++)  <<
				  " ON " << setw(6) << gen_name("t", x->name) << " AT " <<
				  x->deadline * ArgPack::soft_deadline_mul_ << "\n";
			}
		}
	DAG_ENDEACH_NODE

	os << "}\n\n\n";
}

/*===========================================================================*/
void TTdag::push_period() {
// Find the max deadline.
	double max_d = 0.0;

	DAG_FOREACH_NODE(TTnode, x, this)
		max_d = max(max_d, x->deadline);
	DAG_ENDEACH_NODE

	period_ = max(period_, max_d);
}

/*===========================================================================*/
RString TTdag::gen_name(const char *header, int name) const {
	return RString(header) + RString(number_) + RString("_") +
	  RString(name);
}

/*===========================================================================*/
void TTdag::compute_dims() {
    // find out how tall and wide the dag is ...
    // Get node depths, DUMB method [O(n^2) I think]
 
	DAG_FOREACH_NODE(TTnode, n, this)
		n->h = n->w = -1;
	DAG_ENDEACH_NODE
	int mh = 0;
	int change;
	do {
		change = 0;
		DAG_FOREACH_NODE(TTnode, n, this)
			int pl = parents_h(n);
			if( pl+1 > n->h ) {
				n->h = pl+1;
				mh = max(mh, n->h);
				change = 1;
			}
		DAG_ENDEACH_NODE
	} while(change);
	maxh = ++mh;   // mh is height

	// heights assigned, next widths
	int *wlist = new int[mh+1];
	for(int k=0; k < mh; k++ ) wlist[k] = 0;
	int mw = 0;
	DAG_FOREACH_NODE(TTnode, n, this)
		n->w = wlist[n->h]++;
		mw = max(mw, n->w);
	DAG_ENDEACH_NODE
	maxw = ++mw;

	delete[] wlist;

}

/*===========================================================================*/
void TTdag::deadlines() {
	DAG_FOREACH_NODE(TTnode, x, this)
// SIMPLE FOR NOW: PUT DEADLINE ON ALL "TERMINAL PROCESSES"
		if( NumOutArcs(x) == 0 )
			x->deadline = (x->h + 1) * ArgPack::vertex_time_;
		else
			x->deadline = -1;

		x->hard = RGen::gen.flip(ArgPack::prob_hard_deadline_);
	DAG_ENDEACH_NODE
}

/*===========================================================================*/
int TTdag::parents_h(TTnode *n) {
	int rv = -1;
	DAG_FOREACH_INARC(TTarc, arc, n, this)
		rv = max(rv, FromNode(arc)->h);
	DAG_ENDEACH_INARC
	return rv;
}
